package dao.commons;

import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.Hibernate;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.ProjectionList;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;

import dao.commons.criteria.CriteriaHelper;
import dao.commons.criteria.Criterion;
import domain.commons.Preselection;

@Repository
public abstract class GenericDAOImpl<T extends Object> implements GenericDAO<T>
{
  private Class entityClass = (Class)((ParameterizedType)getClass().getGenericSuperclass()).getActualTypeArguments()[0];
 
  @Autowired
  protected SessionFactory sessionFactory;
  
  @Override
  public final T get(int id)
  {
    return (T)sessionFactory.getCurrentSession().get(entityClass, id);
  }
  
  @Override
  public final void refresh(T entity)
  {
	  sessionFactory.getCurrentSession().refresh(entity);
  }

  @Override
  public void save(T entity)
  {
	  sessionFactory.getCurrentSession().save(entity);
  }

  @Override
  public void saveOrUpdate(T entity)
  {
	  sessionFactory.getCurrentSession().saveOrUpdate(entity);
  }

  @Override
  public void update(T entity)
  {
	  sessionFactory.getCurrentSession().update(entity);
  }

  @Override
  public final void delete(T entity)
  {
	  sessionFactory.getCurrentSession().delete(entity);
  }

  @Override
  public void evict(Object entity)
  {
	  sessionFactory.getCurrentSession().evict(entity);
  }

  @Override
  public void persist(Object entity)
  {
	  sessionFactory.getCurrentSession().persist(entity);
  }

  @Override
  public final void flush()
  {
	  sessionFactory.getCurrentSession().flush();
  }

  @Override
  public final void clearSession()
  {
	  sessionFactory.getCurrentSession().clear();
  }

  @Override
  public final List<T> loadAll()
  {
    return null; //sessionFactory.getCurrentSession().loadAll(entityClass);
  }
  
  @Override
  public final boolean exists(int id)
  {
    Criteria criteria = sessionFactory.getCurrentSession().createCriteria(entityClass).add(Restrictions.idEq(id));
    criteria.setProjection(Projections.rowCount());
    return (Integer)criteria.uniqueResult() > 0;
  }

  @Override
  public final void initialize(Object proxy)
  {
    Hibernate.initialize(proxy);
  }

  @Override
  public final List<T> findByCriteria(Preselection preselection)
  {
	  return findByCriteria(preselection, true);
  }
  
  @Override
  public final List<T> findByCriteria(Preselection preselection,boolean ograniczona)
  {
	if (!preselection.isActive())
    {
      return null;
    }

    Criteria criteria = sessionFactory.getCurrentSession().createCriteria(preselection.getEntityClass());

    CriteriaHelper helper = new CriteriaHelper(criteria);

    if (preselection.getCriteriaList() != null)
    {
      for (Criterion c : preselection.getCriteriaList())
      {
        c.prepareCriteria(helper.getSubcriteria(c.getNestedPath()));
      }
    }

    criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
    criteria.setProjection(Projections.countDistinct("id"));

    Long rowCount = (Long) criteria.uniqueResult();
    preselection.setRowCount(rowCount.intValue());

    if (rowCount == 0)
      return null;

    helper.addOrder(preselection.getSortProperty(), preselection.isSortAscending());

    ProjectionList projectionList = Projections.projectionList();

    projectionList.add(Projections.distinct(Projections.id()));
    projectionList.add(CriteriaHelper.getProjectionForSortPropertyPostgresFix(preselection.getSortProperty()));

    criteria.setProjection(projectionList);

    if (ograniczona) 
    {
    	criteria.setFirstResult(preselection.getCurrentPage() * preselection.getPageSize());
    	criteria.setMaxResults(preselection.getPageSize());
    }
    
    List<Integer> idList = new ArrayList<Integer>();

    List<Object> resultList = criteria.list();
    for (Object o : resultList)
    {
      Object[] row = (Object[]) o;
      Integer id = (Integer) row[0];
      idList.add(id);
    }
   
    criteria = sessionFactory.getCurrentSession().createCriteria(preselection.getEntityClass());
    helper = new CriteriaHelper(criteria);

    criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
    criteria.add(Restrictions.in("id", idList));
    
    helper.addOrder(preselection.getSortProperty(), preselection.isSortAscending());

    List list = criteria.list();
    
    for (int i=0;i<list.size();++i) list.set(i, (T)list.get(i));
    return list;
  }
  
  @Override
  public final List<T> findByFirstLetter(List<Criterion> criteriaList)
  {
	  if (criteriaList == null || criteriaList.isEmpty()) return null;
	  	  
	  Criteria criteria = sessionFactory.getCurrentSession().createCriteria(entityClass);
  
	  CriteriaHelper helper = new CriteriaHelper(criteria);

	  for (Criterion c : criteriaList)
	   {
	        c.prepareCriteria(helper.getSubcriteria(c.getNestedPath()));
	   }
	  
	    criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
	    criteria.setProjection(Projections.countDistinct("id"));

	    int rowCount = (Integer) criteria.uniqueResult();
	   
	    if (rowCount == 0) return null;

	    ProjectionList projectionList = Projections.projectionList();

	    projectionList.add(Projections.distinct(Projections.id()));
	    projectionList.add(CriteriaHelper.getProjectionForSortPropertyPostgresFix(criteriaList.get(0).getProperty()));

	    criteria.setProjection(projectionList);

	    //criteria.setMaxResults(150);

	    List<Integer> idList = new ArrayList<Integer>();

	    List<Object> resultList = criteria.list();
	    for (Object o : resultList)
	    {
	      Object[] row = (Object[]) o;
	      Integer id = (Integer) row[0];
	      idList.add(id);
	    }
	   
	    criteria = sessionFactory.getCurrentSession().createCriteria(entityClass);
	    helper = new CriteriaHelper(criteria);

	    criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
	    criteria.add(Restrictions.in("id", idList));
	    helper.addOrder(criteriaList.get(0).getProperty(), true);

	    List list = criteria.list();
	    
	    for (int i=0;i<list.size();++i) list.set(i, (T)list.get(i));
	    return list;
  }
        
  @Override
  public <Entity> Entity getReferencedEntity(Class entityClass, int id, String referencedEntity)
  {
    Criteria c = sessionFactory.getCurrentSession().createCriteria(entityClass).add(Restrictions.idEq(id));
    c.setProjection(Projections.property(referencedEntity));
    
    return (Entity)c.uniqueResult();
  }

  @Override
  public Integer getReferencedEntityId(Class entityClass, int id, String referencedEntity)
  {
    Criteria c = sessionFactory.getCurrentSession().createCriteria(entityClass).add(Restrictions.idEq(id));
    c.createCriteria(referencedEntity, "entity_");

    c.setProjection(Projections.property("entity_.id"));
    
    return (Integer)c.uniqueResult();
  }

public SessionFactory getSessionFactory() {
	return sessionFactory;
}

public void setSessionFactory(SessionFactory sessionFactory) {
	this.sessionFactory = sessionFactory;
}
  
  

  
}
