package service.security;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.sql.Timestamp;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.persistence.ManyToMany;

import org.apache.log4j.Logger;
import org.hibernate.Session;
import org.hibernate.type.Type;
import org.springframework.orm.hibernate3.SessionFactoryUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import commons.BibliotekaThreadContext;
import commons.security.audit.AuditableNumer;
import commons.security.audit.NonAuditableField;
import commons.security.hibernate.acl.DMLOperationType;

import dao.commons.GenericDAOImpl;
import domain.security.AuditRecord;

@Transactional(propagation = Propagation.MANDATORY)
@Service
public class AuditLoggerServiceImpl extends GenericDAOImpl<AuditRecord> implements AuditLoggerService
{

  protected final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

  protected final Logger           logger     = Logger.getLogger(getClass());

  @Override
  public void logDelete(Object entity, Serializable id)
  {
    String numer = null;

    if (entity instanceof AuditableNumer)
      numer = ((AuditableNumer) entity).getNumer();

    auditOperation((Integer) id, entity.getClass().getSimpleName(), numer, DMLOperationType.Delete, "id", null, null);
  }

  @Override
  public void logUpdate(Object entity, Serializable id, Object[] currentState, Object[] previousState, String[] propertyNames,
      Type[] types) throws RuntimeException
  {
    Session session = null;
    try
    {

      String numer = null;

      if (entity instanceof AuditableNumer)
        numer = ((AuditableNumer) entity).getNumer();

      if (previousState == null)
      {
        session = SessionFactoryUtils.getNewSession(getSessionFactory());
        previousState = getPreviousState(session, entity, id, types, previousState, propertyNames);
      }

      for (int i = 0; i < propertyNames.length; i++)
      {
        if (!isAuditable(entity, propertyNames[i]))
        {
          continue;
        }

        if (types[i].isCollectionType())
        {

        }
        else if (types[i].isAssociationType())
        {
          auditOperation((Integer) id, entity.getClass().getSimpleName(), numer, DMLOperationType.Update, propertyNames[i] + "."
              + types[i].getName(), getFKFromState(currentState[i]), getObjectId(previousState[i]));
        }
        else
        {
          auditOperation((Integer) id, entity.getClass().getSimpleName(), numer, DMLOperationType.Update, propertyNames[i],
              currentState[i], previousState[i]);
        }

      }
    }
    catch (IllegalAccessException e)
    {
      throw new RuntimeException(e);
    }
    catch (InvocationTargetException e)
    {
      throw new RuntimeException(e);
    }
    finally
    {
      if (session != null)
      {
        SessionFactoryUtils.releaseSession(session, getSessionFactory());
      }
    }
  }

  public boolean isAuditable(Object entity, String propertyName)
  {
    Field[] fields = getAllFields(entity.getClass(), null);
    for (int j = 0; j < fields.length; j++)
    {
      if (fields[j].getName().equals(propertyName) && fields[j].isAnnotationPresent(NonAuditableField.class))
      {
        return false;
      }
    }

    return true;
  }

  @Override
  public void logSave(Object entity, Serializable id, Object[] state, String[] propertyNames, Type[] types)
      throws SecurityException
  {

    String numer = null;

    if (entity instanceof AuditableNumer)
      numer = ((AuditableNumer) entity).getNumer();

    for (int i = 0; i < propertyNames.length; i++)
    {

      if (!isAuditable(entity, propertyNames[i]))
      {
        continue;
      }

      if (types[i].isCollectionType())
      {
        Field[] fields = getAllFields(entity.getClass(), null);
        for (int j = 0; j < fields.length; j++)
        {
          if (propertyNames[i].equals(fields[j].getName()))
          {
            if (fields[j].isAnnotationPresent(ManyToMany.class))
            {
              ManyToMany mtm = fields[j].getAnnotation(ManyToMany.class);

              if (mtm.mappedBy() == null || mtm.mappedBy().length() == 0)
              {
                logCollection((Integer) id, state[i], entity.getClass().getSimpleName() + "." + propertyNames[i]);
              }
            }
          }
        }
      }
      else if (types[i].isAssociationType())
      {
        auditOperation((Integer) id, entity.getClass().getSimpleName(), numer, DMLOperationType.Insert, propertyNames[i],
            getFKFromState(state[i]), null);
      }
      else
      {
        auditOperation((Integer) id, entity.getClass().getSimpleName(), numer, DMLOperationType.Insert, propertyNames[i], state[i],
            null);
      }
    }
  }

  protected Object[] getPreviousState(Session session, Object entity, Serializable id, Type[] types, Object[] previousState,
      String[] propertyNames) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException
  {

    Object[] prvs = new Object[propertyNames.length];

    Object obj = session.get(entity.getClass(), id);

    Map<String, Object> fNames = new HashMap<String, Object>();

    Field[] declaredFields = obj.getClass().getDeclaredFields();
    for (int i = 0; i < declaredFields.length; i++)
    {
      declaredFields[i].setAccessible(true);
      fNames.put(declaredFields[i].getName(), declaredFields[i].get(obj));
    }

    // Method[] methods = obj.getClass().getMethods();
    for (int i = 0; i < propertyNames.length; i++)
    {
      // Method m = findGetter(methods, propertyNames[i]);
      // Object o = m.invoke(obj, null);

      prvs[i] = fNames.get(propertyNames[i]);
    }

    return prvs;
  }

  private Method findGetter(Method[] methods, String fieldName)
  {
    for (Method m : methods)
    {
      if ((m.getName().startsWith("get") || m.getName().startsWith("is"))
          && m.getName().toLowerCase().endsWith(fieldName.toLowerCase()))
        return m;

    }

   return null;
  }

  private Serializable getObjectId(Object obj)
  {

    if (obj == null)
    {
      return null;
    }

    Class objectClass = obj.getClass();
    Method[] methods = objectClass.getMethods();

    Serializable persistedObjectId = null;
    for (int ii = 0; ii < methods.length; ii++)
    {

      if (methods[ii].getName().equals("getId"))
      {
        try
        {
          persistedObjectId = (Serializable) methods[ii].invoke(obj, null);
          break;
        }
        catch (Exception e)
        {

        }
      }
    }
    return persistedObjectId.toString();
  }

  protected String getFKFromState(Object o)
  {
    if (o == null)
    {
      return null;
    }

    Field[] fields = getAllFields(o.getClass(), null);

    for (int j = 0; j < fields.length; j++)
    {
      try
      {
        fields[j].setAccessible(true);

        if (fields[j].getName().equals("id"))
        {

          try
          {
            fields[j].setAccessible(true);
            Object forKey;
            forKey = fields[j].get(o);

            if (forKey != null)
            {
              return (String) forKey.toString();
            }
          }
          catch (Exception e)
          {
            logger.error(e);
            throw new RuntimeException(e);
          }
        }

      }
      catch (Exception e)
      {

      }
    }

    return null;
  }

  private void logCollection(Integer id, Object collection, String colName)
  {
    if (collection instanceof Set)
    {
      Set set = (Set) collection;
      for (Object o : set)
      {
        Field[] fields = getAllFields(o.getClass(), null);
        for (int j = 0; j < fields.length; j++)
        {
          if (fields[j].getName().equals("id"))
          {
            fields[j].setAccessible(true);
            Object forKey;
            try
            {
              forKey = fields[j].get(o);
              if (forKey != null)
              {
                auditOperation(id, colName, null, DMLOperationType.Update, "fk ", forKey, null);
              }
            }
            catch (Exception e)
            {
              logger.error(e);
              throw new SecurityException(e);
            }

          }
        }
      }
    }
  }

  protected Field[] getAllFields(Class objectClass, Field[] fields)
  {

    Field[] newFields = objectClass.getDeclaredFields();

    int fieldsSize = 0;
    int newFieldsSize = 0;

    if (fields != null)
    {
      fieldsSize = fields.length;
    }

    if (newFields != null)
    {
      newFieldsSize = newFields.length;
    }

    Field[] totalFields = new Field[fieldsSize + newFieldsSize];

    if (fieldsSize > 0)
    {
      System.arraycopy(fields, 0, totalFields, 0, fieldsSize);
    }

    if (newFieldsSize > 0)
    {
      System.arraycopy(newFields, 0, totalFields, fieldsSize, newFieldsSize);
    }

    Class superClass = objectClass.getSuperclass();

    Field[] finalFieldsArray;

    if (superClass != null && !superClass.getName().equals("java.lang.Object"))
    {
      finalFieldsArray = getAllFields(superClass, totalFields);
    }
    else
    {
      finalFieldsArray = totalFields;
    }

    return finalFieldsArray;

  }

  protected void auditOperation(int objectPK, String className, String numer, DMLOperationType dmlOperation, String propName,
      Object newValue, Object oldValue)
  {

    if (oldValue == null && newValue == null)
      return;

    if (oldValue != null && newValue != null && newValue.equals(oldValue))
    {
      return;
    }

    if (propName.equals("id") || propName.equals("optLock") || propName.equals("aclDomena"))
    {
      return;
    }

    if (newValue instanceof BigDecimal)
    {
      newValue = ((BigDecimal) newValue).setScale(2, BigDecimal.ROUND_HALF_DOWN).toEngineeringString();
      if (oldValue != null)
        oldValue = ((BigDecimal) oldValue).setScale(2, BigDecimal.ROUND_HALF_DOWN).toEngineeringString();
    }
    else if (newValue instanceof Date)
    {
      newValue = dateFormat.format(newValue);
      if (oldValue != null)
        oldValue = dateFormat.format(oldValue);
    }
   
    AuditRecord auditRecord = new AuditRecord();
    auditRecord.setTimeStamp(new Timestamp(new Date().getTime()));
    auditRecord.setLogin(BibliotekaThreadContext.getLogin());
    auditRecord.setTransactionID(sessionFactory.getCurrentSession().getTransaction().hashCode());
    auditRecord.setObjectPK(objectPK);
    auditRecord.setPropertyName(propName);
    auditRecord.setClassName(className);
    auditRecord.setDmlOperation(dmlOperation);
    auditRecord.setNumer(numer);

    if (newValue != null)
    {
      auditRecord.setNewValue(newValue.toString());
    }
    if (oldValue != null)
    {
      auditRecord.setOldValue(oldValue.toString());
    }

    save(auditRecord);
  }

}
